#!/bin/bash

set -e

helper_basename="$(basename "$0")"
const_pkg='pkg'
const_xml='testing-result.xml'
const_error='*** ERROR ***'
const_work_dir='testing-work'

opt_clone_base=
opt_requires=
opt_skill=
opt_testing_pkg=
opt_triggers=
opt_nice='<auto>'
opt_report_format='xml'
opt_assume_ready='no'

this_pkg=
pkgs_ready=
preloaded_pkgs=
mk_work_dir=1
pkg_to_this_pkg='../..'

function usage {
    message="$*"

    cat <<EOF
Usage:
  $helper_basename <command> [<args>]

Commands:
  test                    Check out dependencies in work dir and run tests;
  test-in <pkg>           Test local package in directory <pkg>;
  with-rhino <cmd>        Execute command <cmd> within a fresh Rhino wrapper;

Common options:
  --nice <cmd>            Prefix interpreter invocations with <cmd>;
  --clone-base <dir>      Sets directory for cloning packages from to <dir>;
  --require <pkg>         Check out and load package <pkg> before this one;

Options for test, test-in:
  --skill <impl>          Use SKILL implementation <impl> (default: wrapper);
  --testing-pkg <dir>     Use testing package from <dir>;
  --report-format <fmt>   Generate report in format <fmt>;

Options for test-in:
  --assume-ready <yes|no> Assume required packages are ready;

EOF

    if test -n "$message"; then
        echo "$message"
        echo
    fi
}

function die_usage {
    local status="$1"; shift
    usage "$const_error $*"
    exit "$status"
}

function die {
    local status="$1"; shift
    echo
    echo "$const_error $*"
    echo
    exit "$status"
}

function read_pkg_requires {
    local head="$1"; shift

    < "$head" sed -n                            \
        -e '/;[; \t]*Code:/q'                   \
        -e '/;[; \t]*Commentary:/q'             \
        -e 's/;[; \t]*@requires//p'
}

function pkg_requires {
    local pkg="$1"; shift
    local pkg_dir="$1"; shift
    local head="$pkg_dir/$pkg.il"

    if ! test -f "$head"; then
        head="$pkg_dir/$pkg.ils"
    fi

    local requires="$(read_pkg_requires "$head")"
    if test -n "$requires"; then
        echo "$requires"
    else
        echo '/'
    fi
}

function ensure_pkgs_ready {
    if test "$opt_assume_ready" = 'yes'; then
        return
    fi

    local work_dir="$1"; shift
    local pkgs="$1"; shift
    local reqs=''

    for pkg in ${pkgs//,/ }; do
        case "$pkgs_ready" in
            */$pkg/*)
                echo "Package "$pkg" already checked-out; ignoring."
                ;;
            *)
                local pkg_dir="$work_dir/$const_pkg/$pkg"
                git clone "$opt_clone_base/$pkg" "$pkg_dir"
                pkgs_ready="$pkgs_ready$pkg/"
                local prs="$(pkg_requires "$pkg" "$pkg_dir/")"
                if test "$prs" != '/'; then
                    reqs="$reqs,$prs"
                fi
                ;;
        esac
    done

    if test -n "${reqs//,/}"; then
        ensure_pkgs_ready "$work_dir" "$reqs"
    fi
}

function pkg_set_globals {
    local clone_base=$(git config remote.origin.url)
    local this_basename="$(basename "$clone_base")"
    local this_dirname="$(dirname "$clone_base")"

    # Set global package name
    if test -z "$this_pkg"; then
        if test "$this_basename" = '.git'; then
            this_pkg="$(basename "$this_dirname")"
            this_dirname="$(dirname "$this_dirname")"
        else
            this_pkg="${this_basename/.git/}"
        fi
    fi

    # Set clone base if necessary.
    if test -z "$opt_clone_base"; then
        opt_clone_base="$this_dirname"
    fi

    if test "$opt_nice" == '<auto>'; then
        local nice=' '
        if which nice >/dev/null 2>&1; then
            nice=' nice'
        fi
        if which ionice >/dev/null 2>&1; then
            nice="${nice} ionice -c 3"
        fi
        opt_nice="${nice:1}"
    fi
}

function pkg_bootstrap {
    if test -n "$pkgs_ready"; then
        return
    fi

    local work_dir="$1"; shift
    pkg_set_globals

    if test "$mk_work_dir" == 1; then
        mk_work_dir=0
        mkdir "$work_dir/"
        mkdir "$work_dir/$const_pkg/"
    else
        mkdir -p "$work_dir/$const_pkg/"
        rm -f "$work_dir/$const_pkg/$this_pkg"
    fi

    if test -n "$opt_testing_pkg"; then
        test -f "$opt_testing_pkg/testing.ils"
        (cd "$work_dir/$const_pkg/" && ln -s "$opt_testing_pkg" 'testing')
        pkgs_ready="${pkgs_ready}/testing/"
    fi

    # Create self-link.
    (cd "$work_dir/$const_pkg/" && ln -s "$pkg_to_this_pkg" "$this_pkg")
    pkgs_ready="$pkgs_ready/$this_pkg/"
    local pkg_dir="$work_dir/$const_pkg/$this_pkg"
    local prs="$(pkg_requires "$this_pkg" "$pkg_dir")"
    if test "$prs" != '/'; then
        ensure_pkgs_ready "$work_dir" "$prs"
    fi

    if test -z "$opt_skill"; then
        preloaded_pkgs='run testing'
        ensure_pkgs_ready "$work_dir" "$preloaded_pkgs"
        make -C "$work_dir/$const_pkg/run/" wrapper
        opt_skill="$const_pkg/run/wrapper"
        test -x "$work_dir/$opt_skill"
    else
        preloaded_pkgs='testing'
        ensure_pkgs_ready "$work_dir" "$preloaded_pkgs"
    fi
}

function parse_options {
    local cmd="$1"; shift

    while test -n "$1"; do
        case "$cmd,$1" in
            *,--nice)
                opt_nice="$2"
                shift 2
                ;;
            *,--clone-base)
                opt_clone_base="$2"
                shift 2
                ;;
            *,--require)
                opt_requires="$opt_requires,$2"
                shift 2
                ;;
            test*,--skill)
                opt_skill="$2"
                shift 2
                ;;
            test*,--testing-pkg)
                opt_testing_pkg="$2"
                shift 2
                ;;
            test*,--report-format)
                opt_report_format="$2"
                shift 2
                ;;
            test-in,--assume-ready)
                opt_assume_ready="$2"
                shift 2
                ;;
            *)
                die_usage 2 "Bad argument/option for $cmd: $1; tail: $*."
                ;;
        esac
    done
}

function test_in {
    local work_dir="$1"; shift
    local boot_prefixes="$1"; shift
    local test_pkg="$1"; shift
    local requires="$1"; shift

    (
        cd "$work_dir"
        test -f "$const_xml" && rm "$const_xml"

        if test "$this_pkg" = 'veda'; then
            if test "${opt_skill:0:1}" = '/'; then
                export VEDA_SKILL_BIN="$opt_skill"
            else
                export VEDA_SKILL_BIN="$PWD/$opt_skill"
            fi
        fi

        local prefixes_form suffix
        if test -z "$opt_testing_pkg"; then
            prefixes_form="(parseString \"$boot_prefixes\")"
            suffix='/testing/testing.ils'
        else
            prefixes_form="(list \"$opt_testing_pkg\")"
            suffix='/testing.ils'
        fi

        $opt_nice "$opt_skill" <<EOF
(progn
  (exists prefix $prefixes_form
    (let ((boot (strcat prefix "$suffix")))
      (when (isFile boot)
        (load boot))))
  (let ((base_opts (quote (nil
                           chattyp t
                           reportFilename "./$const_xml")))
        (other_opts (list (quote reportFormat) (concat "$opt_report_format")
                          (quote _preloaded) (parseString "$preloaded_pkgs"))))
    (let ((r (VedaTestingRunTestsOnPackage "$test_pkg"
                                           ?options (append base_opts
                                                            other_opts))))
      (if r
          (exit 0)
          (printf "\\n$const_error Test failures occured.\\n")))))
(exit 1)
EOF

        if test "$opt_report_format" = 'xml'; then
            test -f "$const_xml" ||                                    \
                die 3 "Missing test results; package: '$test_pkg'; "    \
                "implementation: '$opt_skill'."
        fi
    )
}

function cmd_test {
    parse_options 'test' "$@"

    local requires="$opt_requires"

    pkg_bootstrap "$const_work_dir"
    ensure_pkgs_ready "$const_work_dir" "$requires"
    test_in "$const_work_dir" "$const_pkg" "$this_pkg" "$requires"
}

function cmd_test_in {
    local work_dir="$1"; shift
    test -n "$work_dir" || \
        die_usage 2 "Command test-in takes a mandatory argument."
    local requires="$opt_requires"
    local resolved_dir="$(cd "$work_dir" && /bin/pwd)"

    this_pkg="$(basename "$resolved_dir")"

    parse_options 'test-in' "$@"

    # KLUDGE: We are changing run directories, so won't benefit from
    # any package checked-out by with-rhino.  This obviously breaks
    # down with test-in ., but we don't care for now.  A better
    # solution would be to use a specific package path.
    pkgs_ready=''

    mk_work_dir=0
    pkg_to_this_pkg='..'
    pkg_bootstrap "$work_dir"
    ensure_pkgs_ready "$work_dir" "$requires"
    test_in "$work_dir" "$const_pkg .." "$this_pkg" "$requires"
}

function cmd_with_rhino {
    local requires="$opt_requires,skill,js"

    pkg_bootstrap "$const_work_dir"
    ensure_pkgs_ready "$const_work_dir" "$requires,rhino"

    local wrapper_dir="$const_work_dir/testing-rhino/"
    local wrapper="$wrapper_dir/wrapper"

    test -f "$wrapper" &&                               \
        die 3 "Wrapper '$wrapper' already present."

    test_in "$const_work_dir" "$const_pkg" 'rhino' "$requires"

    test -x "$wrapper" ||                       \
        die 3 "Wrapper '$wrapper' missing."

    # Continue with fresh Rhino wrapper.
    opt_skill="$(cd "$wrapper_dir" && echo "$PWD/wrapper")"
    preloaded_pkgs='testing'
    main "$@"
}

function main {
    case "$1" in
        test)
            shift
            cmd_test "$@"
            ;;
        test-in)
            shift
            cmd_test_in "$@"
            ;;
        with-rhino)
            shift
            cmd_with_rhino "$@"
            ;;
        *)
            die_usage 2 "Unrecognized command: $*"
            ;;
    esac
}

main "$@"
